#!/usr/bin/env python3
import click
import dill
import os
import importlib
import pandas as pd
from datetime import timedelta, datetime

# Modules from this repository
from .parsers.base_parser import BaseParser
from .plots.base_plot import BasePlot
from glupredkit.helpers.unit_config_manager import unit_config_manager
from glupredkit.helpers.model_config_manager import ModelConfigurationManager, generate_model_configuration
import glupredkit.helpers.cli as helpers


# TODO: Fix so that all default values are defined upstream (=here in the CLI), and removed from downstream


@click.command()
def setup_directories():
    """Set up necessary directories for GluPredKit."""
    cwd = os.getcwd()
    print("Creating directories...")

    folder_path = 'data'
    folder_names = ['raw', 'configurations', 'trained_models', 'figures', 'reports']

    for folder_name in folder_names:
        path = os.path.join(cwd, folder_path, folder_name)
        os.makedirs(path, exist_ok=True)
        print(f"Created directory {path}.")

    print("Directories created for usage of GluPredKit.")


@click.command()
@click.option('--parser', type=click.Choice(['tidepool', 'nightscout', 'apple_health', 'ohio_t1dm']),
              help='Choose a parser', required=True)
@click.option('--username', type=str, required=False)
@click.option('--password', type=str, required=False)
@click.option('--start-date', type=str,
              help='Start date for data retrieval. Default is two weeks ago. Format "dd-mm-yyyy"')
@click.option('--file-path', type=str, required=False)
@click.option('--end-date', type=str, help='End date for data retrieval. Default is now. Format "dd-mm-yyyy"')
@click.option('--output-file-name', type=str, help='The file name for the output.')
def parse(parser, username, password, start_date, file_path, end_date, output_file_name):
    """Parse data and store it as CSV in data/raw using a selected parser"""

    # Load the chosen parser dynamically based on user input
    parser_module = importlib.import_module(f'glupredkit.parsers.{parser}')

    # Ensure the chosen parser inherits from BaseParser
    if not issubclass(parser_module.Parser, BaseParser):
        raise click.ClickException(f"The selected parser '{parser}' must inherit from BaseParser.")

    # Create an instance of the chosen parser
    chosen_parser = parser_module.Parser()

    click.echo(f"Parsing data using {parser}...")

    date_format = "%d-%m-%Y"

    if not end_date:
        end_date = datetime.now()
    else:
        end_date = datetime.strptime(end_date, date_format)

    if not start_date:
        start_date = end_date - timedelta(days=14)
    else:
        start_date = datetime.strptime(start_date, date_format)

    # Ensure that the optional params match the parser
    if parser in ['tidepool', 'nightscout']:
        if username is None or password is None:
            raise ValueError(f"{parser} parser requires that you provide --username and --password") 
        else:
            parsed_data = chosen_parser(start_date, end_date, username=username, password=password)
    elif parser in ['apple_health', 'ohio_t1dm']:
        if file_path is None:
            raise ValueError(f"{parser} parser requires that you provide --file-path")
        else:
            parsed_data = chosen_parser(start_date, end_date, file_path=file_path)
    else:
        raise ValueError("unrecognized parser: '{parser}'")
    
    output_path = 'data/raw/'
    date_format = "%d-%m-%Y"

    if output_file_name:
        file_name = output_file_name + '.csv'
    else:
        file_name = (parser + '_' + start_date.strftime(date_format) + '_to_' + end_date.strftime(date_format) + '.csv')

    click.echo("Storing data as CSV...")
    helpers.store_data_as_csv(parsed_data, output_path, file_name)
    click.echo(f"Data stored as CSV at '{output_path}' as '{file_name}'")
    click.echo(f"Data has the shape: {parsed_data.shape}")


@click.command()
@click.option('--file-name', prompt='Configuration file name', help='Name of the configuration file.')
@click.option('--data', prompt='Input data file name (from data/raw/)', help='Name of the data file from data/raw/.')
@click.option('--preprocessor', prompt='Preprocessor (available: basic)', help='Name of the preprocessor.')
@click.option('--prediction-horizons', prompt='Prediction horizons in minutes (comma-separated without space)',
              help='Comma-separated list of prediction horizons.')
@click.option('--num-lagged-features', prompt='Number of lagged features', help='Number of lagged features.')
@click.option('--num-features', prompt='Numerical features (a subset of column names from the input data file)',
              help='Comma-separated list of numerical features.')
@click.option('--cat-features', prompt='Categorical features (press enter if none)', default='',
              help='Comma-separated list of categorical features.')
@click.option('--test-size', prompt='Test size (float between 0 and 1. Example: 0.25)', help='Test size.')
def generate_config(file_name, data, preprocessor, prediction_horizons, num_lagged_features, num_features, cat_features,
                    test_size):
    prediction_horizons = [int(val) for val in helpers.split_string(prediction_horizons.replace(' ', ''))]
    num_features = helpers.split_string(num_features.replace(' ', ''))
    cat_features = helpers.split_string(cat_features.replace(' ', ''))

    generate_model_configuration(file_name, data, preprocessor, prediction_horizons, int(num_lagged_features),
                                 num_features, cat_features, float(test_size))
    click.echo(f"Storing configuration file to data/configurations/{file_name}...")
    click.echo(f"Note that it might take a minute before the file appears in the folder.")


@click.command()
@click.argument('model', type=click.Choice(['arx',
                                            'elastic_net',
                                            'gradient_boosting_trees',
                                            'huber',
                                            'lasso',
                                            'lstm',
                                            'random_forest',
                                            'ridge',
                                            'svr_linear',
                                            'svr_rbf',
                                            'tcn'
                                            ]))
@click.argument('config-file-name', type=str)
def train_model(model, config_file_name):
    """
    This method does the following:
    1) Process data using the given configurations
    2) Train the given models for the given prediction horizons in the configuration
    """
    click.echo(f"Starting pipeline to train model {model} with configurations in {config_file_name}...")
    model_config_manager = ModelConfigurationManager(config_file_name)

    prediction_horizons = model_config_manager.get_prediction_horizons()
    model_module = helpers.get_model_module(model)

    for prediction_horizon in prediction_horizons:
        # PREPROCESSING

        # Perform data preprocessing using your preprocessor
        train_data, _ = helpers.get_preprocessed_data(prediction_horizon, model_config_manager)
        click.echo(f"Training data finished preprocessing...")

        # MODEL TRAINING
        # Create an instance of the chosen model
        chosen_model = model_module.Model(prediction_horizon)

        processed_data = chosen_model.process_data(train_data, model_config_manager, real_time=False)
        x_train = processed_data.drop('target', axis=1)
        y_train = processed_data['target']

        click.echo(f"Training model...")

        # Initialize and train the model
        model_instance = chosen_model.fit(x_train, y_train)
        click.echo(f"Model {model} with prediction horizon {prediction_horizon} minutes trained successfully!")

        # Assuming model_instance is your class instance
        output_path = "data/trained_models/"
        output_file_name = f'{model}__{config_file_name}__{prediction_horizon}.pkl'
        try:
            with open(f'{output_path}{output_file_name}', 'wb') as f:
                click.echo(f"Saving model {model} to {output_path}{output_file_name}...")
                dill.dump(model_instance, f)
        except Exception as e:
            click.echo(f"Error saving model {model}: {e}")

        if hasattr(model_instance, 'best_params'):
            click.echo(f"Model hyperparameters: {model_instance.best_params()}")


@click.command()
@click.option('--models', help='List of trained models separated by comma, without ".pkl". If none, all '
                               'models will be evaluated. ',
              default=None)
@click.option('--metrics', help='List of metrics to be computed, separated by comma. '
                                'By default all metrics will be computed. ', default='mae,rmse,pcc,parkes_error_grid,clarke_error_grid')
def calculate_metrics(models, metrics):
    """
    This command stores a report of the given metrics in data/reports/.
    """
    trained_models_path = "data/trained_models/"

    if models is None:
        models = helpers.list_files_in_directory(trained_models_path)
    else:
        models = helpers.split_string(models)

    # Prepare a list of metrics
    metrics = helpers.split_string(metrics)

    results = []
    click.echo(f"Calculating metrics...")
    for model_file in models:
        model_name, config_file_name, prediction_horizon = (model_file.split('__')[0], model_file.split('__')[1],
                                                            int(model_file.split('__')[2].split('.')[0]))

        model_config_manager = ModelConfigurationManager(config_file_name)
        model_instance = helpers.get_trained_model(model_file)
        _, test_data = helpers.get_preprocessed_data(prediction_horizon, model_config_manager)

        processed_data = model_instance.process_data(test_data, model_config_manager, real_time=False)
        x_test = processed_data.drop('target', axis=1)
        y_test = processed_data['target']

        y_pred = model_instance.predict(x_test)

        for metric in metrics:
            metric_module = helpers.get_metric_module(metric)
            chosen_metric = metric_module.Metric()
            score = chosen_metric(y_test, y_pred)
            results.append({'Model name': model_name,
                            'Configuration': config_file_name,
                            'Prediction horizon': prediction_horizon,
                            'Metric': metric,
                            'Score': score})

    # Convert results to DataFrame and save as CSV
    df_results = pd.DataFrame(results)
    metric_results_path = 'data/reports/'
    os.makedirs(metric_results_path, exist_ok=True)

    timestamp = datetime.now().isoformat()
    safe_timestamp = timestamp.replace(':', '_')  # Windows does not allow ":" in file names
    results_file_name = f'{safe_timestamp}.csv'

    df_results.to_csv(metric_results_path + results_file_name, index=False)

    click.echo(
        f"{metrics} for trained_models {models} are stored in '{metric_results_path}' as '{results_file_name}'")

    click.echo(df_results.to_string(index=False))


@click.command()
@click.option('--models', help='List of trained models separated by comma, without ".pkl". If none, all '
                               'models will be evaluated. ', default=None)
@click.option('--plots', help='List of plots to be computed, separated by comma. '
                              'By default a scatter plot will be drawn. ', default='scatter_plot')
@click.option('--is-real-time', type=bool, help='Whether to include test data without matching true measurements.'
    , default=False)
@click.option('--start-date', type=str,
              help='Start date for the predictions. Default is the first sample in the test data. '
                   'Format "dd-mm-yyyy/hh:mm"', default=None)
@click.option('--end-date', type=str,
              help='End date, or prediction date for one prediction plots. Default is the last sample in the test data.'
                   'Format "dd-mm-yyyy/hh:mm"', default=None)
@click.option('--carbs', type=int,
              help='Artificial carbohydrate input for one prediction plots. Only available when is-real-time is true.',
              default=None)
@click.option('--insulin', type=float,
              help='Artificial insulin input for one prediction plots. Only available when is-real-time is true.',
              default=None)
def draw_plots(models, plots, is_real_time, start_date, end_date, carbs, insulin):
    """
    This command draws the given plots and store them in data/figures/.

    The "real_time" parameter indicates whether all test data (including when there is no true values) should be
    included. This parameter allows for real-time plots, but is not compatible with all the plots.
    """
    # Prepare a list of plots
    plots = helpers.split_string(plots)

    if models is None:
        models = helpers.list_files_in_directory('data/trained_models/')
    else:
        models = helpers.split_string(models)

    click.echo(f"Calculating predictions...")
    models_data = []
    for model_file in models:
        model_name, config_file_name, prediction_horizon = (model_file.split('__')[0], model_file.split('__')[1],
                                                            int(model_file.split('__')[2].split('.')[0]))

        model_config_manager = ModelConfigurationManager(config_file_name)
        model_instance = helpers.get_trained_model(model_file)
        _, test_data = helpers.get_preprocessed_data(prediction_horizon, model_config_manager, carbs=carbs,
                                                     insulin=insulin, start_date=start_date, end_date=end_date)

        processed_data = model_instance.process_data(test_data, model_config_manager, real_time=is_real_time)
        x_test = processed_data.drop('target', axis=1)
        y_test = processed_data['target']

        y_pred = model_instance.predict(x_test)

        model_data = {
            'name': f'{model_name} {prediction_horizon}-minutes PH',
            'y_pred': y_pred,
            'y_true': y_test,
            'prediction_horizon': prediction_horizon,
            'config': config_file_name,
            'real_time': is_real_time,
            'carbs': carbs,
            'insulin': insulin,
        }
        models_data.append(model_data)

    plot_results_path = 'data/figures/'

    # Draw plots
    click.echo(f"Drawing plots...")
    os.makedirs(plot_results_path, exist_ok=True)
    for plot in plots:
        plot_module = importlib.import_module(f'glupredkit.plots.{plot}')
        if not issubclass(plot_module.Plot, BasePlot):
            raise click.ClickException(f"The selected plot '{plot}' must inherit from BasePlot.")

        chosen_plot = plot_module.Plot()
        chosen_plot(models_data)

    click.echo(f"{plots} for trained_models {models} are stored in '{plot_results_path}'")


@click.command()
@click.option('--use-mgdl', type=bool, help='Set whether to use mg/dL or mmol/L')
def set_unit(use_mgdl):
    # Update the config
    unit_config_manager.use_mgdl = use_mgdl
    click.echo(f'Set unit to {"mg/dL" if use_mgdl else "mmol/L"}.')


# Create a Click group and add the commands to it
cli = click.Group(commands={
    'setup_directories': setup_directories,
    'parse': parse,
    'generate_config': generate_config,
    'train_model': train_model,
    'calculate_metrics': calculate_metrics,
    'draw_plots': draw_plots,
    'set_unit': set_unit,
})

if __name__ == "__main__":
    cli()



"""
This Temporal Convolutional Network (TCN) should be used together with data preprocessed using tf_keras.py.
"""
import numpy as np
import tensorflow as tf
import ast
from tensorflow.keras.layers import LSTM, Dense, Embedding, Flatten, concatenate, Input
from tensorflow.keras.callbacks import EarlyStopping, ReduceLROnPlateau, Callback
from sklearn.model_selection import TimeSeriesSplit
from tcn import TCN
from .base_model import BaseModel
from glupredkit.helpers.tf_keras import process_data


class Model(BaseModel):
    def __init__(self, prediction_horizon):
        super().__init__(prediction_horizon)
        # The recommended approach for saving and loading Keras models is to use Keras's built-in .save() and
        # Using legacy .h5 file type because .keras threw error with M1 Mac chip
        self.model_path = f"data/.keras_models/tcn_ph-{prediction_horizon}.h5"

    def fit(self, x_train, y_train):
        sequences = [np.array(ast.literal_eval(seq_str)) for seq_str in x_train['sequence']]
        targets = y_train.tolist()

        sequences = np.array(sequences)
        targets = np.array(targets)

        # Model architecture
        input_layer = Input(shape=(sequences.shape[1], sequences.shape[2]))

        # TCN layer
        tcn_out = TCN(nb_filters=50, return_sequences=False)(input_layer)

        output_layer = Dense(1)(tcn_out)

        model = tf.keras.Model(inputs=input_layer, outputs=output_layer)
        model.compile(
            optimizer=tf.keras.optimizers.legacy.Adam(learning_rate=0.001, beta_1=0.999, beta_2=0.999, clipnorm=1.0),
            loss='mse')

        # Callbacks
        early_stopping = EarlyStopping(monitor='val_loss', patience=3, restore_best_weights=True)
        reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.1, patience=10, min_lr=0.0001)

        # Split the data into 5 folds
        tscv = TimeSeriesSplit(n_splits=5)

        # Use first 4 folds for training and the last fold for validation
        for train_idx, val_idx in tscv.split(sequences):
            train_X, val_X = sequences[train_idx], sequences[val_idx]
            train_Y, val_Y = targets[train_idx], targets[val_idx]

            model.fit(train_X, train_Y, validation_data=(val_X, val_Y), epochs=20, batch_size=1,
                      callbacks=[early_stopping, reduce_lr])

        model.save(self.model_path)

        return self

    def predict(self, x_test):
        sequences = [np.array(ast.literal_eval(seq_str)) for seq_str in x_test['sequence']]
        sequences = np.array(sequences)

        model = tf.keras.models.load_model(self.model_path, custom_objects={"Adam": tf.keras.optimizers.legacy.Adam,
                                                                            "TCN": TCN})
        predictions = model.predict(sequences)

        return [val[0] for val in predictions]

    def best_params(self):
        # Return the best parameters found by GridSearchCV
        return None

    def process_data(self, df, model_config_manager, real_time):
        return process_data(df, model_config_manager, real_time)



from sklearn.base import BaseEstimator, TransformerMixin
from glupredkit.helpers.model_config_manager import ModelConfigurationManager


class BaseModel(BaseEstimator, TransformerMixin):
    def __init__(self, prediction_horizon):
        self.prediction_horizon = prediction_horizon

    def fit(self, x_train, y_train):
        # Perform any additional processing of the input features here

        # Fit the model
        # ...

        raise NotImplementedError("Model has not implemented fit method!")

    def predict(self, x_test):
        # Perform any additional processing of the input features here

        # Make predictions using the fitted model
        # ...

        # Return the predictions
        raise NotImplementedError("Model has not implemented predict method!")

    def best_params(self):
        # Return the best parameters found by GridSearchCV
        raise NotImplementedError("Model has not implemented predict method!")

    def process_data(self, df, model_config_manager: ModelConfigurationManager, real_time: bool):
        # Implement library specific preprocessing steps that are required before training a pandas dataframe
        raise NotImplementedError("Model has not implemented predict method!")



"""
The Apple Health parser is processing the raw .xml data from Apple Health export and returning the data merged into
the same time grid in a dataframe.
"""
from .base_parser import BaseParser
import xml.etree.ElementTree as ET
import pandas as pd


def get_data_frame_for_type(data, type, name):
    df = data[data.type == type]
    df = df.copy()

    # Converting to mg/dL if necessary
    if name == "CGM":
        unit = df.iloc[0]['unit']
        if str(unit).startswith('mmol'):
            df['value'] = df['value'].apply(lambda x: x * 18.018)

    df.rename(columns={'value': name}, inplace=True)
    df.rename(columns={'startDate': 'date'}, inplace=True)
    df.drop(['type', 'sourceName', 'sourceVersion', 'unit', 'creationDate', 'device', 'endDate'], axis=1, inplace=True)
    df.set_index('date', inplace=True)
    return df


class Parser(BaseParser):
    def __init__(self):
        super().__init__()

    def __call__(self, start_date, end_date, file_path: str):
        tree = ET.parse(file_path)
        root = tree.getroot()
        record_list = [x.attrib for x in root.iter('Record')]
        data = pd.DataFrame(record_list)

        # Dates from string to datetime
        for col in ['creationDate', 'startDate', 'endDate']:
            data[col] = pd.to_datetime(data[col], errors='coerce')

        # value is numeric, NaN if fails
        data['value'] = pd.to_numeric(data['value'], errors='coerce')

        # Filter the data on the input start- and enddate
        data['startDate'] = data['startDate'].dt.tz_localize(None)
        data = data[(data['startDate'] >= start_date) & (data['startDate'] <= end_date)]

        # Creating separate dataframes for each data type
        df_glucose = get_data_frame_for_type(data, 'HKQuantityTypeIdentifierBloodGlucose', 'CGM')
        df_insulin = get_data_frame_for_type(data, 'HKQuantityTypeIdentifierInsulinDelivery', 'insulin')
        df_carbs = get_data_frame_for_type(data, 'HKQuantityTypeIdentifierDietaryCarbohydrates', 'carbs')
        df_heartrate = get_data_frame_for_type(data, 'HKQuantityTypeIdentifierHeartRate', 'heartrate')
        df_heartratevariability = get_data_frame_for_type(data, 'HKQuantityTypeIdentifierHeartRateVariabilitySDNN',
                                                      'heartratevariability')
        df_caloriesburned = get_data_frame_for_type(data, 'HKQuantityTypeIdentifierActiveEnergyBurned', 'caloriesburned')
        df_respiratoryrate = get_data_frame_for_type(data, 'HKQuantityTypeIdentifierRespiratoryRate', 'respiratoryrate')
        df_steps = get_data_frame_for_type(data, 'HKQuantityTypeIdentifierStepCount', 'steps')
        df_restingheartrate = get_data_frame_for_type(data, 'HKQuantityTypeIdentifierRestingHeartRate', 'restingheartrate')

        # TODO: Add workouts and sleep data

        # Resampling all datatypes into the same time-grid
        df = df_glucose.copy()
        df = df.resample('5T', label='right').mean()

        df_carbs = df_carbs.resample('5T', label='right').sum().fillna(value=0)
        df = pd.merge(df, df_carbs, on="date", how='outer')
        df['carbs'] = df['carbs'].fillna(value=0.0)

        df_insulin = df_insulin.resample('5T', label='right').sum().fillna(value=0)
        df = pd.merge(df, df_insulin, on="date", how='outer')
        df['insulin'] = df['insulin'].fillna(value=0.0)

        df_heartrate = df_heartrate.resample('5T', label='right').mean()
        df = pd.merge(df, df_heartrate, on="date", how='outer')

        df_heartratevariability = df_heartratevariability.resample('5T', label='right').mean()
        df = pd.merge(df, df_heartratevariability, on="date", how='outer')

        df_caloriesburned = df_caloriesburned.resample('5T', label='right').sum().fillna(value=0)
        df = pd.merge(df, df_caloriesburned, on="date", how='outer')
        df['caloriesburned'] = df['caloriesburned'].fillna(value=0.0)

        df_respiratoryrate = df_respiratoryrate.resample('5T', label='right').mean()
        df = pd.merge(df, df_respiratoryrate, on="date", how='outer')

        df_steps = df_steps.resample('5T', label='right').sum().fillna(value=0)
        df = pd.merge(df, df_steps, on="date", how='outer')
        df['steps'] = df['steps'].fillna(value=0.0)

        df_restingheartrate = df_restingheartrate.resample('5T', label='right').mean()
        df = pd.merge(df, df_restingheartrate, on="date", how='outer')

        # Add hour of day
        df['hour'] = df.index.hour

        return df



class BasePlot:
    def __init__(self):
        pass

    def __call__(self, models_data):
        """
        Draw a plot for the given trained_models data.

        models_data: A list of dictionaries containing the model name, y_true, and y_pred.
                    Example: [{'name': 'model1', 'y_pred': [...]}, ...]
        y_true: A list of true measured values
        """

        raise NotImplementedError("BasePlot not implemented!")




from .base_plot import BasePlot
import numpy as np
import matplotlib.pyplot as plt
from glupredkit.helpers.unit_config_manager import unit_config_manager


class Plot(BasePlot):
    def __init__(self):
        super().__init__()

    def __call__(self, models_data):
        """
        This plot plots predicted trajectories from the measured values. It is recommended to use around 24 hours
        of test data for this plot to work optimally.

        #TODO: Predict trajectories when the models with different prediction horizons have the same config.
        """

        def on_hover(event):
            if event.inaxes == ax:
                for hover_line in lines:
                    contains, _ = hover_line.contains(event)
                    if contains:
                        hover_line.set_alpha(1.0)
                    else:
                        hover_line.set_alpha(0.2)
                fig.canvas.draw_idle()

        if unit_config_manager.use_mgdl:
            unit = "mg/dL"
        else:
            unit = "mmol/L"

        unique_model_names = set()  # Initialize an empty set to store unique names
        unique_configs = set()
        unique_ph = set()

        real_time = False
        for model_entry in models_data:
            name = model_entry['name'].split(' ')[0]
            real_time = model_entry['real_time']
            config = model_entry['config']
            ph = model_entry['prediction_horizon']
            unique_model_names.add(name)  # Add the name to the set
            unique_configs.add(config)
            unique_ph.add(ph)

        unique_model_names_list = list(unique_model_names)
        unique_configs_list = list(unique_configs)
        unique_ph_list = list(unique_ph)

        max_ph = max(unique_ph_list)
        max_target_index = max_ph // 5

        trajectories = []  # Will be a set of trajectories for models with equal names and configs

        for model_name in unique_model_names_list:
            for config in unique_configs_list:
                filtered_entries = [entry for entry in models_data if
                                    entry['name'].split(' ')[0] == model_name and
                                    entry['config'] == config]

                if len(filtered_entries) != 0:
                    prediction_horizons = [0]
                    y_true = filtered_entries[0]['y_true'].dropna()
                    predictions = [y_true]

                    if not unit_config_manager.use_mgdl:
                        predictions = [[unit_config_manager.convert_value(val) for val in
                                       y_true]]

                    for entry in filtered_entries:
                        entry_prediction_horizon = entry['prediction_horizon']
                        start_index = entry_prediction_horizon // 5
                        prediction_horizons = prediction_horizons + [entry_prediction_horizon]
                        entry_predictions = entry['y_pred'][start_index:]

                        if unit_config_manager.use_mgdl:
                            predictions = predictions + [entry_predictions]
                        else:
                            predictions = predictions + [[unit_config_manager.convert_value(val) for val in
                                                          entry_predictions]]

                    # Sort lists by prediction horizons
                    pairs = list(zip(prediction_horizons, predictions))
                    sorted_pairs = sorted(pairs, key=lambda x: x[0])
                    sorted_prediction_horizons, sorted_predictions = zip(*sorted_pairs)

                    # Add trajectory data for reference value
                    trajectory_data = {
                        'prediction_horizons': list(sorted_prediction_horizons),
                        'predictions': list(sorted_predictions),  # A list of lists of predicted values
                        'y_true': y_true,
                        'model_name': model_name,
                        'config': config,
                    }
                    trajectories.append(trajectory_data)

        for trajectory_data in trajectories:
            fig, ax = plt.subplots()

            model_name = trajectory_data.get('model_name')
            y_prediction_lists = trajectory_data.get('predictions')
            y_true = trajectory_data.get('y_true')
            prediction_horizons = trajectory_data.get('prediction_horizons')

            total_time = len(y_true) * 5 + max_target_index * 5
            t = np.arange(0, total_time, 5)

            # Use correct unit
            if unit_config_manager.use_mgdl:
                ax.axhspan(70, 180, facecolor='blue', alpha=0.2)
            else:
                y_true = [unit_config_manager.convert_value(val) for val in y_true]
                ax.axhspan(unit_config_manager.convert_value(70), unit_config_manager.convert_value(180),
                           facecolor='blue', alpha=0.2)

            ax.set_title('Blood glucose predicted trajectories')
            ax.set_xlabel('Time (minutes)')
            ax.set_ylabel(f'Blood glucose [{unit}]')
            ax.scatter(t[:len(y_true)], y_true, label='Blood glucose measurements', color='black')

            lines = []
            # Add predicted trajectories
            for i in range(len(y_true)):
                trajectory = []
                t_values = []

                for j in range(len(prediction_horizons)):
                    if i < len(y_prediction_lists[j]):
                        t_values = t_values + [i*5 + prediction_horizons[j]]
                        trajectory = trajectory + [y_prediction_lists[j][i]]
                    else:
                        break

                line, = ax.plot(t_values, trajectory, linestyle='--')
                lines.append(line)

            fig.canvas.mpl_connect('motion_notify_event', on_hover)
            ax.legend()
            plt.title(f'Predicted trajectories for {model_name}')
            file_path = "data/figures/"
            # TODO: Add config name
            file_name = f"trajectories_{trajectory_data.get('config')}_{model_name}.png"
            plt.savefig(file_path + file_name)
            plt.show()



from .base_metric import BaseMetric
from error_grids import zone_accuracy


class Metric(BaseMetric):
    def __init__(self):
        super().__init__('Parkes Error Grid')

    def __call__(self, y_true, y_pred):
        accuracy_values = zone_accuracy(y_true, y_pred, 'parkes')
        formatted_values = ["{:.1f}%".format(value * 100) for value in accuracy_values]

        return formatted_values



"""
The Ohio T1DM parser is processing the raw .xml data from the Ohio T1DM datasets and returning the data merged into
the same time grid in a dataframe.
"""
from .base_parser import BaseParser
import xml.etree.ElementTree as ET
import pandas as pd


class Parser(BaseParser):
    def __init__(self):
        super().__init__()

    def __call__(self, start_date, end_date, file_path: str):
        tree = ET.parse(file_path)
        root = tree.getroot()

        dataframes = {}
        for child in root:
            tag_name = child.tag
            events = []
            for event in child.findall('event'):
                events.append(event.attrib)
            dataframes[tag_name] = pd.DataFrame(events)

        # Resampling all datatypes into the same time-grid
        df = dataframes['glucose_level'].copy()

        df['ts'] = pd.to_datetime(df['ts'], format='%d-%m-%Y %H:%M:%S', errors='coerce')
        df['value'] = pd.to_numeric(df['value'], errors='coerce')
        df.rename(columns={'value': 'CGM', 'ts': 'date'}, inplace=True)
        df.set_index('date', inplace=True)
        df = df.resample('5T', label='right').mean()#.ffill(limit=1)

        # Carbohydrates
        df_carbs = dataframes['meal'].copy()
        df_carbs['ts'] = pd.to_datetime(df_carbs['ts'], format='%d-%m-%Y %H:%M:%S', errors='coerce')
        df_carbs['carbs'] = pd.to_numeric(df_carbs['carbs'], errors='coerce')
        df_carbs.rename(columns={'ts': 'date'}, inplace=True)
        df_carbs = df_carbs[['date', 'carbs']]
        df_carbs.set_index('date', inplace=True)
        df_carbs = df_carbs.resample('5T', label='right').sum().fillna(value=0)
        df = pd.merge(df, df_carbs, on="date", how='outer')
        df['carbs'] = df['carbs'].fillna(value=0.0)

        # Bolus doses
        df_bolus = dataframes['bolus'].copy()
        df_bolus['ts_begin'] = pd.to_datetime(df_bolus['ts_begin'], format='%d-%m-%Y %H:%M:%S', errors='coerce')
        df_bolus['dose'] = pd.to_numeric(df_bolus['dose'], errors='coerce')
        df_bolus.rename(columns={'ts_begin': 'date', 'dose': 'bolus'}, inplace=True)
        df_bolus = df_bolus[['date', 'bolus']]
        df_bolus.set_index('date', inplace=True)
        df_bolus = df_bolus.resample('5T', label='right').sum().fillna(value=0)

        # Basal rates
        df_basal = dataframes['basal'].copy()
        df_basal['ts'] = pd.to_datetime(df_basal['ts'], format='%d-%m-%Y %H:%M:%S', errors='coerce')
        df_basal['value'] = pd.to_numeric(df_basal['value'])
        df_basal.rename(columns={'ts': 'date', 'value': 'basal'}, inplace=True)
        df_basal = df_basal[['date', 'basal']]
        df_basal.set_index('date', inplace=True)
        df_basal = df_basal.resample('5T', label='right').asfreq().ffill()

        # Temp basal rates
        df_temp_basal = dataframes['temp_basal'].copy()
        if not df_temp_basal.empty:
            df_temp_basal['ts_begin'] = pd.to_datetime(df_temp_basal['ts_begin'], format='%d-%m-%Y %H:%M:%S',
                                                       errors='coerce')
            df_temp_basal['ts_end'] = pd.to_datetime(df_temp_basal['ts_end'], format='%d-%m-%Y %H:%M:%S',
                                                     errors='coerce')
            # Override the basal rates with the temp basal rate data
            for index, row in df_temp_basal.iterrows():
                start_date = row['ts_begin']  # Assuming the column name in df_temp_basal is 'start_date'
                end_date = row['ts_end']  # Assuming the column name in df_temp_basal is 'end_date'
                value = row['value']
                df_basal.loc[start_date:end_date] = float(value)

        # Convert basal rates from U/hr to U
        df_basal['basal'] = pd.to_numeric(df_basal['basal'], errors='coerce')
        df_basal['basal'] = df_basal['basal'] / 12
        df_basal.rename(columns={'basal': 'insulin'}, inplace=True)
        df_bolus.rename(columns={'bolus': 'insulin'}, inplace=True)

        df_insulin = df_basal.add(df_bolus, fill_value=0)
        df = pd.merge(df, df_insulin, on="date", how='outer')
        df['insulin'] = df['insulin'].fillna(value=0.0)

        # Heart rate
        df_heartrate = dataframes['basis_heart_rate'].copy()
        df_heartrate['ts'] = pd.to_datetime(df_heartrate['ts'], format='%d-%m-%Y %H:%M:%S', errors='coerce')
        df_heartrate['value'] = pd.to_numeric(df_heartrate['value'], errors='coerce')
        df_heartrate.rename(columns={'ts': 'date', 'value': 'heartrate'}, inplace=True)
        df_heartrate.set_index('date', inplace=True)
        df_heartrate = df_heartrate.resample('5T', label='right').mean()
        df = pd.merge(df, df_heartrate, on="date", how='outer')

        return df



"""
The nightscout parser uses nightscout API to fetch some data using user credentials
and return the data in a format that can be used as input to the blood glucose prediction trained_models.
"""
from aiohttp import ClientError, ClientConnectorError, ClientResponseError
import nightscout
from .base_parser import BaseParser
import pandas as pd
import datetime


class Parser(BaseParser):
    def __init__(self):
        super().__init__()

    def __call__(self, start_date, end_date, username: str, password: str):
        """
        In the nighscout parser, the username is the nightscout URL, and the password is the API key.
        """
        try:
            api = nightscout.Api(username, api_secret=password)

            api_start_date = start_date.strftime('%Y-%m-%dT%H:%M:%S.000Z')
            api_end_date = end_date.strftime('%Y-%m-%dT%H:%M:%S.000Z')

            # gte = greater than or equal to, lte = lower than or equal to
            query_params = {'count': 0, 'find[dateString][$gte]': api_start_date,
                            'find[dateString][$lte]': api_end_date}

            # Dataframe Glucose (SGV) [mg/dL]
            entries = api.get_sgvs(query_params)
            dates = [entry.date for entry in entries]
            values = [entry.sgv for entry in entries]
            df_glucose = pd.DataFrame({'date': dates, 'CGM': values})
            df_glucose.sort_values(by='date', inplace=True, ascending=True)
            df_glucose.set_index('date', inplace=True)

            # Treatments call returns insulin (bolus and basal separated) and carbohydrate intakes
            query_params = {'count': 0, 'find[timestamp][$gte]': api_start_date, 'find[timestamp][$lte]': api_end_date}
            treatments = api.get_treatments(query_params)

            # Dataframe Carbs
            dates = []
            values = []
            for treatment in treatments:
                if treatment.eventType == 'Carb Correction':
                    dates.append(treatment.timestamp)
                    values.append(treatment.carbs)
            df_carbs = pd.DataFrame(
                {'date': dates, 'carbs': values})
            df_carbs.sort_values(by='date', inplace=True, ascending=True)
            df_carbs.set_index('date', inplace=True)

            # Dataframe Bolus
            dates = []
            values = []
            for treatment in treatments:
                if 'Bolus' in treatment.eventType:
                    dates.append(treatment.timestamp)
                    values.append(treatment.insulin)
            df_bolus = pd.DataFrame({'date': dates, 'insulin': values})
            df_bolus.sort_values(by='date', inplace=True, ascending=True)
            df_bolus.set_index('date', inplace=True)

            # Dataframe Basal rates [U/hr]
            dates = []
            values = []
            for treatment in treatments:
                if 'Temp Basal' == treatment.eventType:
                    dates.append(treatment.timestamp)
                    values.append(treatment.rate)
                elif 'Basal' in treatment.eventType:
                    dates.append(treatment.timestamp)
                    values.append(treatment.rate)
            df_basal = pd.DataFrame(
                {'date': dates, 'basal_rate': values})
            df_basal.sort_values(by='date', inplace=True, ascending=True)
            df_basal.set_index('date', inplace=True)

            # Resampling all datatypes into the same time-grid
            df = df_glucose.copy()
            df = df.resample('5T', label='right').mean()

            df_carbs = df_carbs.resample('5T', label='right').sum().fillna(value=0)
            df = pd.merge(df, df_carbs, on="date", how='outer')
            df['carbs'] = df['carbs'].fillna(value=0.0)

            df_bolus = df_bolus.resample('5T', label='right').sum()
            df = pd.merge(df, df_bolus, on="date", how='outer')

            # The reason why NS basal data is different from Tidepool and Apple health,
            # Is that the stored data is not based on actually delivered basal units (like in Apple health)
            # Nor is it the derived temperal basal rate from the actually delivered doses (like in Tidepool),
            # But NS is based on the "programmed temperal basal rate (see metadata apple health),
            # Which is unfortunately quite unaccurate compared to the delivered doses from the pump.
            df_basal = df_basal.resample('5T', label='right').last()
            df_basal['basal_rate'] = df_basal['basal_rate'] / 60 * 5  # From U/hr to U (5-minutes)
            df = pd.merge(df, df_basal, on="date", how='outer')
            df['basal_rate'] = df['basal_rate'].ffill(limit=12 * 24 * 2) # Forward fill the basal rate
            df[['insulin', 'basal_rate']] = df[['insulin', 'basal_rate']].fillna(value=0.0)
            df['insulin'] = df['insulin'] + df['basal_rate']
            df.drop(columns=(["basal_rate"]), inplace=True)

            # Get the current datetime in UTC, given the calendar on current computer
            current_timezone = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
            df.index = df.index.tz_convert(current_timezone)

            # Add hour of day
            df['hour'] = df.index.hour

            return df

        except ClientResponseError as error:
            raise RuntimeError("Received ClientResponseError") from error
        except (ClientError, ClientConnectorError, TimeoutError, OSError) as error:
            raise RuntimeError("Received client error or timeout. Make sure that the username (nightscout URL) and "
                               "passoword (API key) is correct.") from error
        except Exception as e:
            raise RuntimeError(f"Error fetching data: {str(e)}")



import unittest
from src.metrics.base_metric import BaseMetric
from src.metrics.rmse import RMSE
from src.metrics.mae import MAE
from src.metrics.van_herpe import VanHerpe
from src.metrics.kovatchev import Kovatchev
from src.metrics.cao import Cao
from src.metrics.bayer import Bayer
from src.metrics.pcc import PCC

class TestMetric(unittest.TestCase):

    def test_base_metric(self):
        # Test that BaseMetric cannot be instantiated
        with self.assertRaises(TypeError):
            metric = BaseMetric()

    def test_rmse_metric(self):
        # Test RMSE using a known input and output
        metric = RMSE()
        input_data = [1, 2, 3, 4, 5]
        target_data = [1, 3, 5, 7, 9]
        expected_output = 2.449489742783178
        output = metric(input_data, target_data)
        self.assertAlmostEqual(output, expected_output)

    def test_pcc_metric(self):
        y_true = [1, 2, 3, 4, 5]
        y_pred = [1, 3, 5, 7, 9]
        expected_output = 1.0
        metric = PCC()
        output = metric(y_true, y_pred)
        self.assertAlmostEqual(output, expected_output)


    def test_mae_metric(self):
        # Test MAE using a known input and output
        metric = MAE()
        input_data = [1, 2, 3, 4, 5]
        target_data = [1, 3, 5, 7, 9]
        expected_output = 2.0
        output = metric(input_data, target_data)
        self.assertAlmostEqual(output, expected_output)

    def test_van_herpe_metric(self):
        metric = VanHerpe()
        input_data = [90]
        target_data = [140]
        expected_output = 34.675294
        output = metric(input_data, target_data)
        self.assertAlmostEqual(output, expected_output, delta=0.0001)

    def test_kovatchev_metric(self):
        metric = Kovatchev()
        input_data = [90]
        target_data = [140]
        expected_output =  10.169543
        output = metric(input_data, target_data)
        self.assertAlmostEqual(output, expected_output, delta=0.0001)
    
    def test_cao_metric(self):
        metric = Cao()
        input_data = [90]
        target_data = [140]
        expected_output = 40.02489
        output = metric(input_data, target_data)
        self.assertAlmostEqual(output, expected_output, delta=0.0001)

    def test_bayer_metric(self):
        metric = Bayer()
        input_data = [90]
        target_data = [140]
        expected_output = 7.71737
        output = metric(input_data, target_data)
        self.assertAlmostEqual(output, expected_output, delta=0.0001)

if __name__ == '__main__':
    unittest.main()


import unittest
from test_metric import TestMetric

if __name__ == '__main__':
    # Create a test suite
    test_suite = unittest.TestSuite()

    # Add tests to the suite
    test_suite.addTest(unittest.makeSuite(TestMetric))

    # Run the tests
    runner = unittest.TextTestRunner()
    runner.run(test_suite)


import json
import pkg_resources
import os


class UnitConfigManager:
    def __init__(self):
        self.default_config_file = pkg_resources.resource_filename('glupredkit', 'unit_config.json')

        self.user_config_dir = os.path.expanduser("~/.glupredkit")
        self.user_config_file = os.path.join(self.user_config_dir, "unit_config.json")  # Fixed the path

        if not os.path.exists(self.user_config_file):
            self._initialize_user_config()

        with open(self.user_config_file, 'r') as f:
            self.config = json.load(f)

    def _initialize_user_config(self):
        os.makedirs(self.user_config_dir, exist_ok=True)
        with open(self.default_config_file, 'r') as src, open(self.user_config_file, 'w') as dst:
            dst.write(src.read())


    @property
    def use_mgdl(self):
        return self.config.get('use_mgdl', True)

    @use_mgdl.setter
    def use_mgdl(self, value):
        self.config['use_mgdl'] = value
        with open(self.user_config_file, 'w') as f:
            json.dump(self.config, f, indent=4)

    def convert_value(self, value):
        if not self.use_mgdl:
            return value / 18.018
        return value


# You can create a global instance of the config manager.
unit_config_manager = UnitConfigManager()



"""
This file defines the schema and provides get methods for the configuration files describing the properties of the
models before preprocessing and training.
"""

import json
import os


def generate_model_configuration(file_name, data, preprocessor, prediction_horizons, num_lagged_features, num_features,
                                 cat_features, test_size):
    # Check if the 'data' string is a valid file in the 'data/raw/' folder
    data_path = os.path.join('data/raw', data)
    if not os.path.isfile(data_path):
        raise ValueError(f"Data file '{data}' not found in 'data/raw/' folder.")

    # Check if 'preprocessor' is a valid preprocessor module
    preprocessor_module = f'glupredkit.preprocessors.{preprocessor}'
    try:
        __import__(preprocessor_module)
    except ImportError:
        raise ValueError(f"Preprocessor '{preprocessor}' not found in 'preprocessors' module.")

    # Check if 'test_size' is a float between 0 and 1
    if not isinstance(test_size, float) or test_size < 0 or test_size > 1:
        raise ValueError("Test size must be a float between 0 and 1.")

    config = {
        "data": data,
        "preprocessor": preprocessor,
        "prediction_horizons": prediction_horizons,
        "num_lagged_features": num_lagged_features,
        "num_features": num_features,
        "cat_features": cat_features,
        "test_size": test_size
    }
    # Save the generated config to a JSON file
    with open(f'data/configurations/{file_name}.json', 'w') as f:
        json.dump(config, f, indent=4)


class ModelConfigurationManager:
    def __init__(self, config_file):
        self.config_file = 'data/configurations/' + config_file + '.json'
        self.schema = {
            "data": str,
            "preprocessor": str,
            "prediction_horizons": list,
            "num_lagged_features": int,
            "num_features": list,
            "cat_features": list,
            "test_size": float
        }
        self.config = self.load_config()

    def load_config(self):
        try:
            with open(self.config_file, 'r') as f:
                config = json.load(f)
                self.validate_config(config)
                return config
        except FileNotFoundError:
            raise FileNotFoundError(f"Config file '{self.config_file}' not found.")

    def validate_config(self, config):
        for key, value_type in self.schema.items():
            if key not in config:
                raise ValueError(f"Missing key '{key}' in the config file.")
            if not isinstance(config[key], value_type):
                raise ValueError(f"Invalid value type for '{key}' in the config file. Expected {value_type}.")

    def get_data(self):
        return self.config["data"]

    def get_preprocessor(self):
        return self.config["preprocessor"]

    def get_prediction_horizons(self):
        return self.config["prediction_horizons"]

    def get_num_lagged_features(self):
        return self.config["num_lagged_features"]

    def get_num_features(self):
        return self.config["num_features"]

    def get_cat_features(self):
        return self.config["cat_features"]

    def get_test_size(self):
        return self.config["test_size"]



"""
The basic preprocessor does the following:
- Standardize numerical features
- One hot encode categorical features
- Add the target value
"""
from .base_preprocessor import BasePreprocessor
import pandas as pd
from sklearn.preprocessing import OneHotEncoder, StandardScaler


class Preprocessor(BasePreprocessor):
    def __init__(self, numerical_features, categorical_features, prediction_horizon, num_lagged_features, test_size):
        super().__init__(numerical_features, categorical_features, prediction_horizon, num_lagged_features, test_size)

    def __call__(self, df):

        # Drop columns that are not included
        df = df[self.numerical_features + self.categorical_features]

        # Add target column
        target_index = self.prediction_horizon // 5
        if self.prediction_horizon % 5 != 0:
            raise ValueError("Prediction horizon must be divisible by 5.")
        df = df.copy()

        df.loc[:, 'target'] = df['CGM'].shift(-target_index)

        # Interpolation using forward fill
        df[self.numerical_features] = df[self.numerical_features].ffill()

        # Transform columns
        if self.numerical_features:
            scaler = StandardScaler()
            df.loc[:, self.numerical_features] = scaler.fit_transform(df.loc[:, self.numerical_features])
        if self.categorical_features:
            encoder = OneHotEncoder(drop='first')  # dropping the first column to avoid dummy variable trap
            encoded_cols = encoder.fit_transform(df[self.categorical_features])
            encoded_df = pd.DataFrame(encoded_cols.toarray(),
                                      columns=encoder.get_feature_names_out(self.categorical_features), index=df.index)
            df = df.drop(columns=self.categorical_features)
            df = pd.concat([df, encoded_df], axis=1)

        # Train and test split
        # Adding a margin of 24 hours to the train and the test data to avoid memory leak
        margin = int((12 * 24 + self.num_lagged_features + target_index) / 2)
        split_index = int((len(df) - margin) * (1 - self.test_size))

        # Split the data into train and test sets
        train_data = df[:split_index - margin]
        test_data = df[split_index + margin:]

        return train_data, test_data



"""
The tidepool parser uses tidepool API to fetch some data using user credentials
and return the data in a format that can be used as input to the blood glucose prediction trained_models.
"""
from tidepool_data_science_project.makedata.tidepool_api import TidepoolAPI
import datetime
import pandas as pd
from dateutil import parser
from .base_parser import BaseParser


class Parser(BaseParser):
    def __init__(self):
        super().__init__

    def __call__(self, start_date, end_date, username: str, password: str):
        """
        Tidepool API ignores time of day in the dates and will always fetch all data from a specific date
        """
        try:
            tp_api = TidepoolAPI(username, password)
            tp_api.login()

            # All the data in json format
            user_data = tp_api.get_user_event_data(start_date, end_date)

            tp_api.logout()

            # Sort data into one dataframe each data type, and retrieve only the essential information
            (glucose_data, bolus_data, basal_data, carb_data, workout_data) = self.parse_json(user_data)

            # Dataframe blood glucose
            df_glucose = pd.json_normalize(glucose_data)[['time', 'value']]
            if not df_glucose.empty:  # If blood glucose values in mmol/L, convert to mg/dL
                if pd.json_normalize(glucose_data)[['units']].loc[0, 'units'] == 'mmol/L':
                    df_glucose['value'] = df_glucose['value'] * 18.0182
            df_glucose.time = df_glucose.time.apply(self.custom_date_parser)
            df_glucose.time = pd.to_datetime(df_glucose.time, errors='coerce')
            df_glucose.rename(columns={"time": "date", "value": "CGM"}, inplace=True)
            df_glucose.sort_values(by='date', inplace=True, ascending=True)
            df_glucose.set_index('date', inplace=True)

            # Dataframe bolus doses
            df_bolus = pd.json_normalize(bolus_data)[['time', 'normal']]
            df_bolus.time = pd.to_datetime(df_bolus.time, errors='coerce')
            df_bolus.rename(columns={"time": "date", "normal": "insulin"}, inplace=True)
            df_bolus.sort_values(by='date', inplace=True, ascending=True)
            df_bolus.set_index('date', inplace=True)

            # Dataframe basal rates
            df_basal = pd.json_normalize(basal_data)[
                ['time', 'rate']]
            df_basal.time = pd.to_datetime(df_basal.time, errors='coerce')
            df_basal.rename(columns={"time": "date", "rate": "basal_rate"}, inplace=True)
            df_basal.sort_values(by='date', inplace=True, ascending=True)
            df_basal.set_index('date', inplace=True)

            # Dataframe carbohydrates
            df_carbs = pd.json_normalize(carb_data)[['time', 'nutrition.carbohydrate.net']]
            df_carbs.time = pd.to_datetime(df_carbs.time, errors='coerce')
            df_carbs.rename(columns={"time": "date", "nutrition.carbohydrate.net": "carbs"}, inplace=True)
            df_carbs.sort_values(by='date', inplace=True, ascending=True)
            df_carbs.set_index('date', inplace=True)

            if not len(workout_data) == 0:
                df_workouts = pd.json_normalize(workout_data)[['time', 'duration.value', 'name']]
                df_workouts.rename(columns={"duration.value": "duration[s]"}, inplace=True)
                df_workouts['name'] = df_workouts['name'].apply(lambda x: x.split()[0])
                df_workouts.time = pd.to_datetime(df_workouts.time, errors='coerce')
            else:
                df_workouts = pd.DataFrame()

            # Resampling all datatypes into the same time-grid
            df = df_glucose.copy()
            df = df.resample('5T', label='right').mean()

            df_carbs = df_carbs.resample('5T', label='right').sum().fillna(value=0)
            df = pd.merge(df, df_carbs, on="date", how='outer')
            df['carbs'] = df['carbs'].fillna(value=0.0)

            df_bolus = df_bolus.resample('5T', label='right').sum()
            df = pd.merge(df, df_bolus, on="date", how='outer')

            df_basal = df_basal.resample('5T', label='right').last()
            df_basal['basal_rate'] = df_basal['basal_rate'] / 60 * 5  # From U/hr to U (5-minutes)
            df = pd.merge(df, df_basal, on="date", how='outer')
            df['basal_rate'] = df['basal_rate'].ffill(limit=12 * 24 * 2)
            df[['insulin', 'basal_rate']] = df[['insulin', 'basal_rate']].fillna(value=0.0)
            df['insulin'] = df['insulin'] + df['basal_rate']
            df.drop(columns=(["basal_rate"]), inplace=True)

            # Add hour of day
            df['hour'] = df.index.hour

            if not df_workouts.empty:
                # Add activity states
                df['activity_state'] = "None"
                df_workouts.apply(lambda x: self.add_activity_states(x['time'], x['duration[s]'], x['name'], df),
                                  axis=1)

            # Get the current datetime in UTC, given the calendar on current computer
            current_timezone = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
            df.index = df.index.tz_convert(current_timezone)

            return df

        except Exception as e:
            raise RuntimeError(f"Error fetching data: {str(e)}")

    def parse_json(self, user_data):
        """
        Sort the user data from tidepool API into lists of the different data types
        Arguments:
        user_data -- the dictionary from an API call to the Tidepool data
        Output:
        Lists of the different data types in the json(glucose_data, bolus_data, basal_data, carb_data)
        """
        glucose_data = []
        bolus_data = []
        basal_data = []
        carb_data = []
        workout_data = []

        # Sort data types into lists
        for data in user_data:
            if data['type'] == 'cbg':
                glucose_data.append(data)
            elif data['type'] == 'bolus':
                bolus_data.append(data)
            elif data['type'] == 'basal':
                basal_data.append(data)
            elif data['type'] == 'food':
                carb_data.append(data)
            elif data['type'] == 'physicalActivity':
                workout_data.append(data)
            else:
                print('Unknown type is not supported: ', data['type'])

        return glucose_data, bolus_data, basal_data, carb_data, workout_data

    def add_activity_states(self, start_date, duration, workout_name, df):
        end_date = start_date + pd.to_timedelta(duration, unit='s')
        df['activity_state'][df.index.to_series().between(start_date, end_date)] = workout_name

    # Your custom parser function
    def custom_date_parser(self, date_str):
        try:
            # Try parsing with the first format
            return parser.parse(date_str)
        except ValueError:
            # If the first format fails, try the second format
            try:
                return parser.parse(date_str, fuzzy=True)
            except ValueError:
                # If second format also fails, return NaT
                return pd.NaT




import numpy as np
from .base_metric import BaseMetric


class Metric(BaseMetric):
    def __init__(self):
        super().__init__("PCC")

    def _calculate_metric(self, y_true, y_pred):
        corr_coef = np.corrcoef(y_true, y_pred)[0, 1]
        return corr_coef



import matplotlib.pyplot as plt
import itertools
import os
from datetime import datetime
from .base_plot import BasePlot
from glupredkit.helpers.unit_config_manager import unit_config_manager


class Plot(BasePlot):
    def __init__(self):
        super().__init__()

    def __call__(self, models_data):
        """
        Plots the scatter plot for the given trained_models data.

        models_data: A list of dictionaries containing the model name, y_true, and y_pred.
                    Example: [{'name': 'model1', 'y_pred': [...]}, ...]
        """
        # Define unique markers for scatter plot
        markers = itertools.cycle(('o', 's', '^', 'v', 'D', '<', '>'))

        if unit_config_manager.use_mgdl:
            unit = "mg/dL"
            max_val = 400
        else:
            unit = "mmol/L"
            max_val = unit_config_manager.convert_value(400)

        plt.figure(figsize=(10, 8))

        for model_data in models_data:
            model_name = model_data.get('name')

            if unit_config_manager.use_mgdl:
                y_pred = model_data.get('y_pred')
                y_true = model_data.get('y_true')
            else:
                y_pred = [unit_config_manager.convert_value(val) for val in model_data.get('y_pred')]
                y_true = [unit_config_manager.convert_value(val) for val in model_data.get('y_true')]

            marker = next(markers)

            plt.scatter(y_true, y_pred, label=model_name, marker=marker, alpha=0.5)

        # Plotting the line x=y
        plt.plot([0, max_val], [0, max_val], 'k-')

        plt.xlabel(f"True Blood Glucose [{unit}]")
        plt.ylabel(f"Predicted Blood Glucose [{unit}]")
        plt.title(f"Scatter Plot of Prediction Accuracy")
        plt.legend(loc='upper left')

        file_path = "data/figures/"
        os.makedirs(file_path, exist_ok=True)

        timestamp = datetime.now().isoformat()
        safe_timestamp = timestamp.replace(':', '_')  # Windows does not allow ":" in file names
        file_name = f'scatter_plot_{safe_timestamp}.png'
        plt.savefig(file_path + file_name)
        plt.show()



from typing import List


class BaseMetric:
    def __init__(self, name):
        self.name = name

    def __call__(self, y_true: List[float], y_pred: List[float]) -> float:
        if len(y_true) != len(y_pred):
            raise ValueError("y_true and y_pred must have the same length")
        try:
            return self._calculate_metric(y_true, y_pred)
        except NotImplementedError:
            raise NotImplementedError("Metric not implemented!")

    def _calculate_metric(self, y_true: List[float], y_pred: List[float]) -> float:
        raise NotImplementedError("Metric not implemented!")

    def __repr__(self):
        return self.name



from .base_metric import BaseMetric
import numpy as np
from glupredkit.helpers.unit_config_manager import unit_config_manager


class Metric(BaseMetric):
    def __init__(self):
        super().__init__('MAE')

    def _calculate_metric(self, y_true, y_pred):
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)

        mae = np.mean(np.abs(y_true - y_pred))

        if unit_config_manager.use_mgdl:
            return mae
        else:
            return unit_config_manager.convert_value(mae)


"""
All preprocessors must:
- Handle feature addition
- Handle feature imputation
- Add a target column named "target"
- Drop nan-values
- Split data into train- and testdata
"""

class BasePreprocessor:
    def __init__(self, numerical_features, categorical_features, prediction_horizon, num_lagged_features, test_size):
        """
        Args:
            prediction_horizon (int): The prediction horizon in minutes.
            num_lagged_features (int): The number of time-lagged features to generate (12 samples corresponds to one
            hour).
            test_size (float): The fraction of data to reserve for testing.
            include_hour (bool): Whether to include the hour of the day as a feature.
        """
        self.numerical_features = numerical_features
        self.categorical_features = categorical_features
        self.prediction_horizon = prediction_horizon
        self.num_lagged_features = num_lagged_features
        self.test_size = test_size

    def __call__(self, data, **kwargs):
        """
        Args:
            data (DataFrame): The input dataset.

        Returns:
            train_data: The dataset for model training.
            test_data: The dataset for model testing.
        """
        raise NotImplementedError("Preprocessor not implemented!")

    def __repr__(self):
        return self.__class__.__name__



import pandas as pd
import sys
import os
import dill
import importlib
from ..models.base_model import BaseModel
from ..metrics.base_metric import BaseMetric
from ..helpers.model_config_manager import ModelConfigurationManager


def read_data_from_csv(input_path, file_name):
    file_path = input_path + file_name
    return pd.read_csv(file_path, index_col="date", parse_dates=True)


def store_data_as_csv(df, output_path, file_name):
    file_path = output_path + file_name
    df.to_csv(file_path)


def split_string(input_string):
    return [] if not input_string else input_string.split(',')


def user_input_prompt(text):
    # Prompt the user for confirmation
    user_response = input(f"{text} (Y/n): ")

    # Convert the user's response to lowercase and check against 'y' and 'n'
    if user_response.lower() == 'y':
        # Continue processing
        print("Continuing processing...")
    elif user_response.lower() == 'n':
        # Stop or exit
        print("Operation aborted by the user.")
        sys.exit()
    else:
        print("Invalid input. Please respond with 'Y' or 'n'.")


def get_metric_module(metric):
    metric_module = importlib.import_module(f'glupredkit.metrics.{metric}')
    # Ensure the chosen parser inherits from BaseParser
    if not issubclass(metric_module.Metric, BaseMetric):
        raise Exception(f"The selected metric '{metric}' must inherit from BaseMetric.")

    return metric_module


def get_model_module(model):
    model_module = importlib.import_module(f'glupredkit.models.{model}')
    # Ensure the chosen parser inherits from BaseParser
    if not issubclass(model_module.Model, BaseModel):
        raise Exception(f"The selected model '{model}' must inherit from BaseModel.")

    return model_module


def get_trained_model(model_file_name):
    model_path = "data/trained_models/"
    with open(model_path + model_file_name, 'rb') as f:
        model_instance = dill.load(f)

    return model_instance


def get_preprocessed_data(prediction_horizon: int, config_manager: ModelConfigurationManager, carbs=None, insulin=None,
                          start_date=None, end_date=None):
    preprocessor = config_manager.get_preprocessor()
    input_file_name = config_manager.get_data()

    print(f"Preprocessing data using {preprocessor} from file data/raw/{input_file_name}, with a prediction "
          f"horizon of {prediction_horizon} minutes...")
    preprocessor_module = importlib.import_module(f'glupredkit.preprocessors.{preprocessor}')
    chosen_preprocessor = preprocessor_module.Preprocessor(config_manager.get_num_features(),
                                                           config_manager.get_cat_features(),
                                                           prediction_horizon, config_manager.get_num_lagged_features(),
                                                           config_manager.get_test_size())
    # Load the input CSV file into a DataFrame
    data = read_data_from_csv("data/raw/", input_file_name)

    if carbs:
        if 'carbs' in data.columns:
            data.at[data.index[-1], 'carbs'] = carbs
        else:
            raise Exception("No input feature named 'carbs'.")

    if insulin:
        if 'insulin' in data.columns:
            data.at[data.index[-1], 'insulin'] = insulin
        else:
            raise Exception("No input feature named 'insulin'.")

    if end_date:
        # Convert the prediction_date string to a datetime object
        end_date = pd.to_datetime(end_date, format="%d-%m-%Y/%H:%M")
        end_date = end_date.tz_localize(data.index.tz)
        nearest_index = abs(data.index - end_date).argmin()
        data = data.iloc[:nearest_index + 1]

    train_data, test_data = chosen_preprocessor(data)

    if start_date:
        # Convert the prediction_date string to a datetime object
        start_date = pd.to_datetime(start_date, format="%d-%m-%Y/%H:%M")
        start_date = start_date.tz_localize(test_data.index.tz)
        nearest_index = abs(test_data.index - start_date).argmin()
        test_data = test_data.iloc[nearest_index:]

    return train_data, test_data


def list_files_in_directory(directory_path):
    file_list = []
    for filename in os.listdir(directory_path):
        if os.path.isfile(os.path.join(directory_path, filename)):
            file_list.append(filename)
    return file_list



import matplotlib.pyplot as plt
import os
import numpy as np
from datetime import datetime, timedelta
from .base_plot import BasePlot
from glupredkit.helpers.unit_config_manager import unit_config_manager


class Plot(BasePlot):
    def __init__(self):
        super().__init__()

    def __call__(self, models_data):
        """
        Plots the scatter plot for the given trained_models data. This plot automatically use the last data for
        plotting.

        models_data: A list of dictionaries containing the model name, y_true, y_pred, prediction horizon and the name
        of the configuration file.
        """
        history_length = 18  # Number of samples of measured CGM values
        sample_rate = 5  # CGM values are sampled every 5 minutes

        t = np.arange(-history_length * sample_rate + sample_rate, 0 + sample_rate, sample_rate)

        unique_model_names = set()  # Initialize an empty set to store unique names
        unique_configs = set()
        unique_ph = set()

        real_time = False
        for model_entry in models_data:
            name = model_entry['name'].split(' ')[0]
            real_time = model_entry['real_time']
            config = model_entry['config']
            ph = model_entry['prediction_horizon']
            unique_model_names.add(name)  # Add the name to the set
            unique_configs.add(config)
            unique_ph.add(ph)

        unique_model_names_list = list(unique_model_names)
        unique_configs_list = list(unique_configs)
        unique_ph_list = list(unique_ph)

        max_ph = max(unique_ph_list)
        max_target_index = max_ph // sample_rate

        trajectories = []  # Will be a set of trajectories for models with equal names and configs

        for model_name in unique_model_names_list:
            for config in unique_configs_list:
                filtered_entries = [entry for entry in models_data if
                                    entry['name'].split(' ')[0] == model_name and
                                    entry['config'] == config]

                if len(filtered_entries) != 0:
                    prediction_horizons = [0]

                    if real_time:
                        predictions = [filtered_entries[0]['y_true'].dropna().iloc[-1]]
                    else:
                        predictions = [filtered_entries[0]['y_true'].iloc[-1 - max_target_index]]

                    for entry in filtered_entries:
                        ph = entry['prediction_horizon']

                        if real_time:
                            prediction = [entry['y_pred'][-1]]
                        else:
                            target_index = ph // sample_rate
                            prediction = [entry['y_pred'][-target_index]]

                        prediction_horizons = prediction_horizons + [ph]
                        predictions = predictions + prediction

                    if not unit_config_manager.use_mgdl:
                        predictions = [unit_config_manager.convert_value(val) for val in predictions]

                    # Sort lists by prediction horizons
                    pairs = list(zip(prediction_horizons, predictions))
                    sorted_pairs = sorted(pairs, key=lambda x: x[0])
                    sorted_prediction_horizons, sorted_predictions = zip(*sorted_pairs)

                    # Add trajectory data for reference value
                    trajectory_data = {
                        'prediction_horizons': list(sorted_prediction_horizons),
                        'predictions': list(sorted_predictions),
                        'model_name': model_name,
                    }
                    trajectories.append(trajectory_data)

        if unit_config_manager.use_mgdl:
            unit = "mg/dL"
        else:
            unit = "mmol/L"

        # TODO: Add potential carbs and insulin input (DRAW IT AS WELL!)

        plt.figure(figsize=(10, 8))

        for trajectory in trajectories:
            plt.plot(trajectory.get('prediction_horizons'), trajectory.get('predictions'),
                     label=f"Predictions for {trajectory.get('model_name')}")

        # This iteration will break after the first iteration, using a random sample of model data
        for model_data in models_data:
            y_true = model_data.get('y_true')
            prediction_horizon = model_data.get('prediction_horizon')

            if real_time:
                last_y_true_values = y_true.dropna()[-history_length:]
            else:
                last_y_true_values = y_true[-history_length - max_target_index:-max_target_index]

            # Get the datetime where t=0. The index/date of y_true is the date of the predicted value
            time_of_prediction = last_y_true_values.index[-1] + timedelta(minutes=prediction_horizon)

            # User correct unit. Add a text label above t=0
            if unit_config_manager.use_mgdl:
                plt.text(0, 5, f'{time_of_prediction}', fontsize=12, ha='center')
                if model_data.get('carbs'):
                    plt.text(0, 20, f"{model_data.get('carbs')}g of carbohydrates", fontsize=12, ha='center')
                if model_data.get('insulin'):
                    plt.text(0, 30, f"{model_data.get('insulin')}U of insulin", fontsize=12, ha='center')
            else:
                plt.text(0, unit_config_manager.convert_value(5), f'{time_of_prediction}', fontsize=12, ha='center')
                last_y_true_values = [unit_config_manager.convert_value(val) for val in last_y_true_values]
                if model_data.get('carbs'):
                    plt.text(0, unit_config_manager.convert_value(20), f"{model_data.get('carbs')}g of carbohydrates",
                             fontsize=12, ha='center')
                if model_data.get('insulin'):
                    plt.text(0, unit_config_manager.convert_value(30), f"{model_data.get('insulin')}U of insulin",
                             fontsize=12, ha='center')

            plt.scatter(t, last_y_true_values, label=f'Blood glucose measurements', color='black')
            break

        if unit_config_manager.use_mgdl:
            plt.axhspan(70, 180, facecolor='blue', alpha=0.2)
        else:
            plt.axhspan(unit_config_manager.convert_value(70), unit_config_manager.convert_value(180),
                        facecolor='blue', alpha=0.2)

        # Get current y-axis limits
        y_min, y_max = plt.ylim()

        # Set new y-axis limits to include 0
        plt.ylim(min(0, int(y_min)), y_max)

        plt.xlabel(f"Time [minutes]")
        plt.ylabel(f"Blood glucose [{unit}]")
        plt.title(f"Single Prediction Plot")
        plt.legend(loc='upper left')

        file_path = "data/figures/"
        os.makedirs(file_path, exist_ok=True)

        timestamp = datetime.now().isoformat()
        safe_timestamp = timestamp.replace(':', '_')  # Windows does not allow ":" in file names
        file_name = f'one_prediction_{safe_timestamp}.png'
        plt.savefig(file_path + file_name)
        plt.show()



import pandas as pd
from glupredkit.helpers.model_config_manager import ModelConfigurationManager


def add_time_lagged_features(col_name, df, num_lagged_features):
    indexes = list(range(1, num_lagged_features + 1))

    for i in indexes:
        new_col_name = col_name + "_" + str(i * 5)
        df = pd.concat([df, df[col_name].shift(i).rename(new_col_name)], axis=1)
    return df


def process_data(df, model_config_manager: ModelConfigurationManager, real_time=False):

    # Add time-lagged features
    for col in model_config_manager.get_num_features():
        df = add_time_lagged_features(col, df, model_config_manager.get_num_lagged_features())

    if real_time:
        df = df.dropna(subset=df.columns.difference(['target']))
    else:
        df = df.dropna()

    return df




"""
This is a preprocessor that is aiming to reproduce the steps described in the following paper:
https://pubmed.ncbi.nlm.nih.gov/32091990/

NOTE: THIS PREPROCESSOR IS NOT WORKING AS INTENDED! Hence, it is not added to the CLI.
"""

from .base_preprocessor import BasePreprocessor
import pandas as pd
from sklearn.preprocessing import OneHotEncoder, Normalizer


# TODO: ADD, IF NOT TRAINING DATA, DONT INCLUDE SAMPLES ADDED BY IMPUTATION
class Preprocessor(BasePreprocessor):
    def __init__(self, numerical_features, categorical_features, prediction_horizon, num_lagged_features, test_size):
        super().__init__(numerical_features, categorical_features, prediction_horizon, num_lagged_features, test_size)

    def __call__(self, df):

        # Drop columns that are not included
        df = df[self.numerical_features + self.categorical_features]

        # Check if any numerical features have NaN values before imputation
        df['imputed'] = df[self.numerical_features].isna().any(axis=1)

        # Interpolation using forward fill
        df[self.numerical_features] = df[self.numerical_features].interpolate(method='polynomial', order=2)

        # Add target column
        target_index = self.prediction_horizon // 5
        if self.prediction_horizon % 5 != 0:
            raise ValueError("Prediction horizon must be divisible by 5.")
        df = df.copy()
        df.loc[:, 'target'] = df['CGM'].shift(-target_index)

        df = df.dropna()

        # Transform columns
        if self.numerical_features:
            scaler = Normalizer()
            df.loc[:, self.numerical_features] = scaler.fit_transform(df.loc[:, self.numerical_features])
        if self.categorical_features:
            encoder = OneHotEncoder(drop='first')  # dropping the first column to avoid dummy variable trap
            encoded_cols = encoder.fit_transform(df[self.categorical_features])
            encoded_df = pd.DataFrame(encoded_cols.toarray(),
                                      columns=encoder.get_feature_names_out(self.categorical_features), index=df.index)
            df = df.drop(columns=self.categorical_features)
            df = pd.concat([df, encoded_df], axis=1)

        # Since the Ohio T1DM dataset is split into train and test data in the folder structure, users must
        # update their configurations to use the correct data instead of splitting inherently in the preprocessor
        return df, df




from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline
from sklearn.linear_model import Lasso
from .base_model import BaseModel
from glupredkit.helpers.scikit_learn import process_data


class Model(BaseModel):
    def __init__(self, prediction_horizon):
        super().__init__(prediction_horizon)

        self.model = None

    def fit(self, x_train, y_train):

        # Define the model
        pipeline = Pipeline([('regressor', Lasso(tol=1))])

        # Define the parameter grid
        param_grid = {
            'regressor__alpha': [0.0001]
        }

        # Define GridSearchCV
        self.model = GridSearchCV(pipeline, param_grid, cv=5, scoring='neg_mean_squared_error')
        self.model.fit(x_train, y_train)
        return self

    def predict(self, x_test):
        # Use the best estimator found by GridSearchCV to make predictions
        y_pred = self.model.best_estimator_.predict(x_test)
        return y_pred

    def best_params(self):
        # Return the best parameters found by GridSearchCV
        return self.model.best_params_

    def process_data(self, df, model_config_manager, real_time):
        return process_data(df, model_config_manager, real_time)



import pandas as pd
import numpy as np
from glupredkit.helpers.model_config_manager import ModelConfigurationManager


def prepare_sequences(data, labels, window_size, step_size=1):
    X, y, dates = [], [], np.array(labels.index)[0:len(data) - window_size]

    for i in range(0, len(data) - window_size, step_size):
        X.append(data[i:i + window_size])
        y.append(labels.iloc[i + window_size])

    return np.array(X), np.array(y), dates


def create_dataframe(sequences, targets, dates):
    # Create a new DataFrame to store sequences and targets

    # Convert sequences to lists
    sequences_as_strings = [str(list(map(list, seq))) for seq in sequences]

    dataset_df = pd.DataFrame({
        'date': dates,
        'sequence': sequences_as_strings,  # sequences_as_lists,
        'target': targets
    })
    dataset_df.set_index('date')
    return dataset_df


def process_data(df, model_config_manager: ModelConfigurationManager, real_time=False):

    if real_time:
        df = df.dropna(subset=df.columns.difference(['target']))
    else:
        df = df.dropna()

    df_X, df_y = df.drop("target", axis=1), df["target"]

    # Add sliding windows of features
    sequences, targets, dates = prepare_sequences(df_X, df_y,
                                                  window_size=model_config_manager.get_num_lagged_features())

    # Store as a dataframe with two columns: targets and sequences
    df = create_dataframe(sequences, targets, dates)

    return df




from setuptools import setup, find_packages

project_name = "GluPredKit"
version = "0.1.5"
author = "Miriam K. Wolff"
author_email = "miriamkwolff@outlook.com"
package_name = "glupredkit"  # The package name on pip install

with open('README.md', encoding='utf-8') as f:
    readme = f.read()

setup(
    name=project_name,
    version=version,
    author=author,
    author_email=author_email,
    packages=find_packages(),
    package_dir={package_name: package_name},
    long_description=readme,
    long_description_content_type="text/markdown",
    python_requires='>=3.7, <3.10',
    license='MIT',
    include_package_data=True,
    install_requires=[
        "matplotlib",
        "pandas",
        "scikit-learn",
        "xgboost",
        "tensorflow",
        "keras-tcn",
        "numpy",
        "aiohttp",
        "click",
        "dill",
        "tidepool-data-science-project",
        "python-nightscout",
        "error-grids",
    ],
    package_data={
        'glupredkit': ['unit_config.json'],
    },
    entry_points={
        'console_scripts': [
            'glupredkit = glupredkit.cli:cli',
        ],
    }
)




class BaseParser:
    def __init__(self):
        pass

    def __call__(self, start_date, end_date, **kwargs):
        """
        start_date -- start date for dataset
        end_date -- end date for dataset
        optional additional parameters such as api username and key

        Returns four dataframes on the format specified in the README.
        """
        raise NotImplementedError("Metric not implemented!")

    def __repr__(self):
        return self.__class__.__name__
